import re
from pathlib import Path
from typing import List

from config_classes import SimpleConfig as Config

import cordage
from cordage import Experiment


def test_dict_return_value(global_config):
    def func(config: Config, cordage_trial):  # noqa: ARG001
        return {"a": 1, "b": "string", "c": Path(".")}

    trial = cordage.run(func, args=[], global_config=global_config)

    metadata_path = trial.output_dir / "cordage.json"

    assert metadata_path.exists()

    experiment = Experiment.from_path(metadata_path)
    metadata = experiment.metadata

    return_value = metadata.result

    assert len(return_value) == 3
    assert "a" in return_value
    assert "b" in return_value
    assert "c" in return_value

    assert return_value["a"] == 1
    assert return_value["b"] == "string"
    assert return_value["c"] == "."


def test_path_return_value(global_config):
    def func(config: Config, cordage_trial):  # noqa: ARG001
        return {"a": 1, "b": "string", "c": Path(".")}

    trial = cordage.run(func, args=[], global_config=global_config)

    metadata_path = trial.output_dir / "cordage.json"

    assert metadata_path.exists()

    experiment = Experiment.from_path(metadata_path)
    metadata = experiment.metadata

    return_value = metadata.result

    assert len(return_value) == 3
    assert "a" in return_value
    assert "b" in return_value
    assert "c" in return_value

    assert return_value["a"] == 1
    assert return_value["b"] == "string"
    assert return_value["c"] == "."


def test_float_return_value(global_config):
    trial_store: List[cordage.Trial] = []

    def func(config: Config, cordage_trial, trial_store=trial_store):  # noqa: ARG001
        return 0.0

    trial = cordage.run(func, args=[], global_config=global_config)

    metadata_path = trial.output_dir / "cordage.json"

    assert metadata_path.exists()

    experiment = Experiment.from_path(metadata_path)
    metadata = experiment.metadata

    assert isinstance(metadata.result, float)
    assert metadata.result == 0.0


def test_unserializable_return_value(global_config, capsys):
    class SomeObject:
        pass

    def func(config: Config, cordage_trial):  # noqa: ARG001
        return SomeObject()

    trial = cordage.run(func, args=[], global_config=global_config)

    captured = capsys.readouterr()

    metadata_path = trial.output_dir / "cordage.json"

    assert metadata_path.exists()

    experiment = Experiment.from_path(metadata_path)
    metadata = experiment.metadata

    assert metadata.result is None

    pattern = r"^.+WARNING.+Cannot serialize .+SomeObject object"

    assert re.search(pattern, captured.err) is not None



from datetime import datetime

import pytest
from config_classes import NestedConfig, SimpleConfig

import cordage
from cordage import Experiment, Series, Trial
from cordage.util import logger


@pytest.mark.timeout(1)
def test_metadata_loading_config_class_casting(global_config):
    def func(config: SimpleConfig):  # noqa: ARG001
        pass

    trial = cordage.run(func, args=["--a", "1", "--b", "2"], global_config=global_config)

    metadata_path = trial.output_dir / "cordage.json"

    assert metadata_path.exists()

    # try loading as dict
    experiment = Experiment.from_path(metadata_path)
    assert isinstance(experiment, Trial)
    assert isinstance(experiment.config, dict)
    assert experiment.config["a"] == 1
    assert experiment.config["b"] == "2"

    # try loading with config class
    experiment = Experiment.from_path(metadata_path, config_cls=SimpleConfig)
    assert isinstance(experiment, Trial)
    assert isinstance(experiment.config, SimpleConfig)
    assert experiment.config.a == 1
    assert experiment.config.b == "2"


def test_trial_series_loading(global_config, resources_path, capsys):
    def func(config: NestedConfig, cordage_trial: cordage.Trial):
        cordage_trial.add_tag(config.alpha.b)

        logger.warning("Trial with alpha.b=%s", config.alpha.b)

    config_file = resources_path / "series_list.yml"

    cordage.run(func, args=[str(config_file)], global_config=global_config)

    series = Experiment.all_from_path(global_config.base_output_dir)[0]

    assert isinstance(series, Series)

    trial_store = [trial.synchronize() for trial in series]

    assert len(trial_store) == 3

    assert all(isinstance(trial, Trial) for trial in trial_store)

    # test log stream
    captured = capsys.readouterr()

    i = 1

    for captured_line in captured.err.strip().split("\n"):
        assert f"Trial with alpha.b=b{i-1}" not in captured_line

        if f"Trial with alpha.b=b{i}" in captured_line:
            i += 1

    assert i == 4

    # after loading the series trials, the configs are merely nested dictionaries
    for i, trial in enumerate(trial_store):
        assert trial.config["alpha"]["b"] == f"b{i+1}"
        assert trial.has_tag(f"b{i+1}")

        assert isinstance(trial.metadata.start_time, datetime)

        # test logging was performed correctly
        assert trial.log_path.exists()

        with trial.log_path.open("r") as fp:
            log_lines = list(fp)

            for j in range(3):
                expected_log_partial = f"Trial with alpha.b=b{j+1}"

                if i == j:
                    assert any(expected_log_partial in line for line in log_lines)
                else:
                    assert not any(expected_log_partial in line for line in log_lines)


def test_trial_series_loading_with_config_class(global_config, resources_path):
    def func(config: NestedConfig, cordage_trial: cordage.Trial):
        cordage_trial.add_tag(config.alpha.b)

        logger.warning("Trial with alpha.b=%s", config.alpha.b)

    config_file = resources_path / "series_list.yml"

    series = cordage.run(func, args=[str(config_file)], global_config=global_config)

    output_dir = series.output_dir

    series = Experiment.from_path(output_dir, config_cls=NestedConfig)

    assert isinstance(series, Series)

    trial_store = [trial.synchronize() for trial in series]

    assert len(trial_store) == 3

    assert all(isinstance(trial, Trial) for trial in trial_store)

    # after loading the series trials, the configs are merely nested dictionaries
    for i, trial in enumerate(trial_store):
        assert isinstance(trial.config, NestedConfig)
        assert trial.config.alpha.b == f"b{i+1}"



from dataclasses import dataclass, field

import cordage


@dataclass
class TrainConfig:
    lr: float = 5e-5


@dataclass
class DataConfig:
    name: str = "MNIST"


@dataclass
class Config:
    data: DataConfig
    training: TrainConfig = field(default_factory=TrainConfig)


def train(config: Config):
    # to something
    print(config)  # noqa: T201


if __name__ == "__main__":
    cordage.run(train)



import logging

import pytest
from config_classes import SimpleConfig as Config

import cordage

log = logging.getLogger(__name__)


def test_exception_logging(global_config):
    """If an (uncaught) exception is thrown in the experiment, it should be logged and noted in the metadata."""

    class SomeSpecificError(RuntimeError):
        pass

    def func(config: Config):  # noqa: ARG001
        msg = "Exception42"
        raise SomeSpecificError(msg)

    context = cordage.FunctionContext(func, global_config=global_config)
    trial = context.parse_args([])

    with pytest.raises(SomeSpecificError):
        context.execute(trial)

    assert trial.has_status("failed")
    assert "exception" in trial.metadata.additional_info
    assert "Exception42" in trial.metadata.additional_info["exception"]["short"]
    assert "Exception42" in trial.metadata.additional_info["exception"]["traceback"]

    with open(trial.log_path) as f:
        log_content = f.read()

    assert "Exception42" in log_content


def test_function_without_annotation():
    def func(config):  # noqa: ARG001
        pass

    with pytest.raises(TypeError) as e_info:
        cordage.run(func, args=[])

    assert "Configuration class could not be derived" in str(e_info.value)


def test_function_without_config_parameter():
    def func():
        pass

    with pytest.raises(TypeError) as e_info:
        cordage.run(func, args=[])

    assert "Callable must accept config" in str(e_info.value)


def test_function_invalid_object_to_execute(global_config):
    def func(config: Config):  # noqa: ARG001
        pass

    context = cordage.FunctionContext(func, global_config=global_config)

    with pytest.raises(TypeError) as e_info:
        context.execute(object())  # type: ignore

    assert "Passed object must be Trial or Series" in str(e_info.value)


def test_multiple_runtime_exceptions(global_config):
    metadata: cordage.Metadata = cordage.Metadata(function="no_function", global_config=global_config)

    with pytest.raises(TypeError):
        exp = cordage.Experiment(metadata, global_config=global_config)

    exp = cordage.Experiment(function="no_function", global_config=global_config)

    with pytest.raises(RuntimeError):
        log.info(str(exp.output_dir))

    assert "status: pending" in repr(exp)



from pathlib import Path

import pytest

from cordage import GlobalConfig
from cordage.util import from_dict as config_from_dict


@pytest.fixture
def global_config(tmp_path: Path) -> GlobalConfig:
    return config_from_dict(
        GlobalConfig,
        {
            "output_dir_format": "experiment{collision_suffix}",
            "base_output_dir": tmp_path / "results",
        },
    )


@pytest.fixture(scope="module")
def resources_path():
    return Path(__file__).parent / "resources"



from pathlib import Path

import pytest
from config_classes import LongConfig as Config

import cordage


def test_simple_config(global_config):
    def func(config: Config):
        """short_function_description

        long_description

        :param config: Configuration to use.
        """
        assert config.a == 1
        assert isinstance(config.b, Path)
        assert config.c == "test"

    cordage.run(func, args=["--a", "1", "--b", "~"], global_config=global_config)


def test_loading(global_config, resources_path):
    def func(config: Config):
        assert config.a == 1
        assert isinstance(config.b, Path)
        assert config.c == "some_other_value"

    config_file = resources_path / "simple_a.toml"

    cordage.run(func, args=[str(config_file), "--c", "some_other_value"], global_config=global_config)


def test_literal_fields(global_config, resources_path):
    def func(config: Config):  # noqa: ARG001
        pass

    config_file = resources_path / "simple_b.json"

    with pytest.raises(cordage.WrongTypeError):
        cordage.run(func, args=[str(config_file)], global_config=global_config)


def test_tuple_length_fields(global_config, resources_path):
    def func(config: Config):  # noqa: ARG001
        pass

    config_file = resources_path / "simple_c.toml"

    with pytest.raises(cordage.WrongTypeError):
        cordage.run(func, args=[str(config_file)], global_config=global_config)


@pytest.mark.skip(reason="dacite currently does not properly support mixed tuples")
def test_valid_mixed_tuple(global_config, resources_path):
    def func(config: Config):  # noqa: ARG001
        pass

    config_file = resources_path / "simple_d.json"

    cordage.run(func, args=[str(config_file)], global_config=global_config)


@pytest.mark.skip(reason="dacite currently does not properly support mixed tuples")
def test_invalid_mixed_tuple(global_config, resources_path):
    def func(config: Config):  # noqa: ARG001
        pass

    config_file = resources_path / "simple_e.toml"

    with pytest.raises(cordage.WrongTypeError):
        cordage.run(func, args=[str(config_file)], global_config=global_config)


def test_valid_optional(global_config, resources_path):
    def func(config: Config):  # noqa: ARG001
        pass

    config_file = resources_path / "simple_f.json"

    cordage.run(func, args=[str(config_file)], global_config=global_config)



import logging

import pytest
from config_classes import AlphaConfig
from config_classes import NestedConfig as Config

import cordage
from cordage import Experiment

log = logging.getLogger(__name__)


def test_trial_series_list(global_config, resources_path):
    global_config.output_dir_format = "nested/structure/experiment{collision_suffix}"

    series_path = global_config.base_output_dir / "nested" / "structure" / "experiment"

    def load_filtered(status=(), tag=()):
        return [exp for exp in Experiment.all_from_path(series_path) if exp.has_status(*status) and exp.has_tag(*tag)]

    def func(config: Config, cordage_trial: cordage.Trial):
        if config.alpha.a == 1:
            cordage_trial.add_tag("the_first")

        if config.alpha.a == 2:
            cordage_trial.add_tag("second")

            assert len(load_filtered()) == 2
            assert len(load_filtered(status=["complete"])) == 1
            assert len(load_filtered(status=["complete", "running"])) == 2

            cordage_trial.comment = "This is #not_the_first(?) run."

        if config.alpha.a == 3:
            cordage_trial.add_tag("not_the_first")

            raise RuntimeError()

    config_file = resources_path / "series_list.yml"

    with pytest.raises(RuntimeError):
        cordage.run(func, args=[str(config_file)], global_config=global_config)

    all_experiments = Experiment.all_from_path(series_path)

    # experiments are sorted
    assert (all_experiments[0].output_dir < all_experiments[1].output_dir) and (
        all_experiments[1].output_dir < all_experiments[2].output_dir
    )

    assert len(load_filtered()) == 3
    assert len(load_filtered(status=["complete"])) == 2
    assert len(load_filtered(tag=["not_the_first"])) == 2
    assert len(load_filtered(tag=["not_the_first", "the_first"], status=["complete"])) == 2
    assert len(load_filtered(tag=["not_the_first"], status=["complete"])) == 1

    # Since we only load top level experiments, only the series should show up
    assert len(Experiment.all_from_path(global_config.base_output_dir)) == 1


def test_annotation_comment(global_config):
    test_comment = "Some string"

    def func(config: Config, cordage_trial):  # noqa: ARG001
        log.info(str(cordage_trial.annotations))

    cordage.run(
        func,
        args=["--alpha.a", "1", "--cordage-comment", test_comment],
        global_config=global_config,
    )

    exp = Experiment.from_path(global_config.base_output_dir / "experiment")
    assert exp.comment == test_comment
    assert exp.annotations["comment"] == test_comment


def test_annotation_comment_addition(global_config, resources_path):
    test_comment = "Some string"
    expected_comment = f"Config file comment\n\n{test_comment}"

    def func(config: Config, cordage_trial):  # noqa: ARG001
        log.info(str(cordage_trial.annotations))

    conf_path = resources_path / "annotation.yaml"

    cordage.run(func, args=["--cordage-comment", test_comment, str(conf_path)], global_config=global_config)

    exp = Experiment.from_path(global_config.base_output_dir / "experiment")
    assert exp.comment == expected_comment
    assert exp.annotations["comment"] == expected_comment


def test_config_annotation_comment(global_config, resources_path):
    expected_comment = "Config file comment"

    def func(config: Config, cordage_trial):  # noqa: ARG001
        log.info(str(cordage_trial.annotations))

    conf_path = resources_path / "annotation.yaml"

    cordage.run(func, args=[str(conf_path)], global_config=global_config)

    exp = Experiment.from_path(global_config.base_output_dir / "experiment")
    assert exp.comment == expected_comment
    assert exp.annotations["comment"] == expected_comment


def test_function_name_saving(global_config, resources_path):
    def func(config: Config):  # noqa: ARG001
        pass

    conf_path = resources_path / "annotation.yaml"

    cordage.run(func, args=[str(conf_path)], global_config=global_config)


def test_aborted_trial(global_config):
    def func(config: AlphaConfig):
        if config.a == 1:
            raise KeyboardInterrupt()

    context = cordage.FunctionContext(func, global_config=global_config)

    trial = context.parse_args(["--a", "2"])
    context.execute(trial)
    assert trial.status == "complete"

    with pytest.raises(KeyboardInterrupt):
        trial = context.parse_args(["--a", "1"])
        context.execute(trial)
    assert trial.status == "aborted"



from dataclasses import replace
from os import PathLike
from typing import Callable, Dict, List, Optional, Type, Union

from dacite.exceptions import (
    DaciteError,
    DaciteFieldError,
    ForwardReferenceError,
    MissingValueError,
    StrictUnionMatchError,
    UnexpectedDataError,
    UnionMatchError,
    WrongTypeError,
)

from cordage.context import FunctionContext
from cordage.experiment import Experiment, Metadata, Series, Trial
from cordage.global_config import GlobalConfig


def run(
    func: Callable,
    args: Optional[List[str]] = None,
    description: Optional[str] = None,
    config_cls: Optional[Type] = None,
    global_config: Union[PathLike, Dict, GlobalConfig, None] = None,
    **kw,
) -> Experiment:
    context = FunctionContext(
        func,
        description=description,
        config_cls=config_cls,
        global_config=replace(GlobalConfig.resolve(global_config), **kw),
    )
    experiment = context.parse_args(args)
    context.execute(experiment)
    return experiment


__all__ = [
    "run",
    "FunctionContext",
    "Experiment",
    "Trial",
    "Metadata",
    "GlobalConfig",
    "Series",
    "DaciteError",
    "DaciteFieldError",
    "WrongTypeError",
    "MissingValueError",
    "UnionMatchError",
    "StrictUnionMatchError",
    "ForwardReferenceError",
    "UnexpectedDataError",
]



import json
import logging
import re
import shutil
from copy import deepcopy
from dataclasses import dataclass, field
from dataclasses import replace as dataclass_replace
from datetime import datetime, timezone
from itertools import count, product
from json.decoder import JSONDecodeError
from math import ceil, floor, log10
from os import PathLike, getpid
from pathlib import Path
from traceback import format_exception
from typing import (
    Any,
    Dict,
    Generator,
    Generic,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from dacite import DaciteError

try:
    import colorlog
except ImportError:
    colorlog = None  # type: ignore

from cordage.global_config import GlobalConfig
from cordage.util import flattened_items, from_dict, logger, nest_items, nested_update, to_dict

T = TypeVar("T")


@dataclass
class Metadata:
    function: str

    global_config: GlobalConfig

    output_dir: Optional[Path] = None
    status: str = "pending"

    start_time: Optional[datetime] = None
    end_time: Optional[datetime] = None

    configuration: Any = None

    result: Any = None

    parent_dir: Optional[Path] = None

    additional_info: Dict = field(default_factory=dict)

    @property
    def duration(self):
        assert self.end_time is not None and self.start_time is not None

        return self.end_time - self.start_time

    def replace(self, **changes):
        return dataclass_replace(self, **changes)

    @property
    def is_series(self):
        return isinstance(self.configuration, dict) and "series_spec" in self.configuration

    def to_dict(self):
        return to_dict(self)

    @classmethod
    def from_dict(cls, data: Mapping):
        return from_dict(cls, data)


class MetadataStore:
    def __init__(self, metadata: Optional[Metadata] = None, /, global_config: Optional[GlobalConfig] = None, **kw):
        self.metadata: Metadata

        if metadata is not None:
            if global_config is not None or len(kw) > 0:
                msg = "Using the `metadata` argument is incompatible with using other arguments."
                raise TypeError(msg)
            else:
                self.metadata = metadata
        else:
            if global_config is None:
                global_config = GlobalConfig()

            self.metadata = Metadata(global_config=global_config, **kw)

    @property
    def global_config(self) -> GlobalConfig:
        return self.metadata.global_config

    @property
    def output_dir(self) -> Path:
        if self.metadata.output_dir is None:
            msg = f"{self.__class__.__name__} has not been started yet."
            raise RuntimeError(msg)
        else:
            return self.metadata.output_dir

    @property
    def parent_dir(self) -> Optional[Path]:
        return self.metadata.parent_dir

    def create_output_dir(self):
        if self.metadata.output_dir is not None:
            self.output_dir.mkdir(parents=True, exist_ok=True)
            return self.output_dir

        tried_paths: Set[Path] = set()
        suffix = ""

        for i in count(1):
            if i > 1:
                level = floor(log10(i) / 2) + 1
                suffix = "_" * level + str(i).zfill(2 * level)

            path = self.global_config.base_output_dir / self.global_config.output_dir_format.format(
                **self.metadata.__dict__,
                collision_suffix=suffix,
            )

            if path in tried_paths:
                # suffix was already tried: assume that further tries wont resolve this collision
                msg = f"Path {path} does already exist - collision could not be avoided."
                raise RuntimeError(msg)

            try:
                path.mkdir(parents=True, exist_ok=False)
                self.metadata.output_dir = path
                return path
            except FileExistsError:
                if self.global_config.overwrite_existing:
                    logger.warning("Path %s does existing. Replacing directory with new one.", str(path))
                    shutil.rmtree(path)
                    path.mkdir(parents=True)
                    self.metadata.output_dir = path
                    return path
                else:
                    tried_paths.add(path)

    @property
    def metadata_path(self):
        return self.output_dir / "cordage.json"

    def save_metadata(self):
        md_dict = self.metadata.to_dict()

        with open(self.metadata_path, "w", encoding="utf-8") as fp:

            def invalid_obj_default(obj):
                logger.warning("Cannot serialize %s", str(obj))

            json.dump(md_dict, fp, indent=4, default=invalid_obj_default)

    @classmethod
    def load_metadata(cls, path: PathLike) -> Metadata:
        path = Path(path)
        if not path.suffix == ".json":
            path = path / "cordage.json"

        with path.open("r", encoding="utf-8") as fp:
            metadata = Metadata.from_dict(json.load(fp))

        if metadata.output_dir != path.parent:
            logger.info(
                f"Output dir is not correct anymore. Changing it to the actual directory"
                f"({metadata.output_dir} -> {path.parent})"
            )
            metadata.output_dir = path.parent

        return metadata


class Annotatable(MetadataStore):
    TAG_PATTERN = re.compile(r"\B#(\w*[a-zA-Z]+\w*)")

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)

        self.annotations = {}

    @property
    def tags(self):
        tags = set(self.explicit_tags)

        # implicit tags
        tags.update(re.findall(self.TAG_PATTERN, self.comment))

        return list(tags)

    @property
    def explicit_tags(self):
        if "tags" not in self.annotations:
            self.annotations["tags"] = []
        return self.annotations["tags"]

    def add_tag(self, *tags: Iterable):
        for t in tags:
            if t not in self.explicit_tags:
                self.explicit_tags.append(t)

    def has_tag(self, *tags: str):
        return len(tags) == 0 or any(t in tags for t in self.tags)

    @property
    def comment(self):
        return self.annotations.get("comment", "") or ""

    @comment.setter
    def comment(self, value):
        self.annotations["comment"] = value

    @property
    def annotations_path(self):
        return self.output_dir / "annotations.json"

    def save_annotations(self):
        with open(self.annotations_path, "w", encoding="utf-8") as fp:
            json.dump(self.annotations, fp, indent=4)

    def load_annotations(self):
        if self.annotations_path.exists():
            with self.annotations_path.open("r", encoding="utf-8") as fp:
                self.annotations = json.load(fp)


class Experiment(Annotatable):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.log_handlers: List[logging.Handler] = []

    def __repr__(self):
        if self.metadata.output_dir is not None:
            return f"{self.__class__.__name__} ({self.output_dir}, status: {self.status})"
        else:
            return f"{self.__class__.__name__} (status: {self.status})"

    @property
    def log_path(self):
        return self.output_dir / self.global_config.logging.filename

    @property
    def status(self) -> str:
        return self.metadata.status

    @property
    def result(self) -> Any:
        return self.metadata.result

    def has_status(self, *status: str):
        return len(status) == 0 or self.status in status

    def start(self):
        """Start the execution of an experiment.

        Set start time, create output directory, registers run, etc.
        """
        assert self.status == "pending", f"{self.__class__.__name__} has already been started."
        self.metadata.start_time = datetime.now(timezone.utc).astimezone()
        self.metadata.status = "running"
        self.metadata.additional_info["process_id"] = getpid()
        self.create_output_dir()
        self.save_metadata()
        self.save_annotations()
        self.setup_log()

    def end(self, status: str = "undecided"):
        """End the execution of an experiment.

        Write metadata, close logs, etc.
        """
        self.metadata.end_time = datetime.now(timezone.utc).astimezone()
        self.metadata.status = status
        self.save_metadata()
        self.save_annotations()
        self.teardown_log()

    def handle_exception(self, exc_type, exc_value, traceback):
        traceback_string = "".join(format_exception(exc_type, value=exc_value, tb=traceback))

        logger.exception("", exc_info=(exc_type, exc_value, traceback))
        self.metadata.additional_info["exception"] = {"short": repr(exc_value), "traceback": traceback_string}

    def __enter__(self):
        self.start()
        logger.info("%s '%s' started.", self.__class__.__name__, str(self.output_dir))

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            logger.info("%s '%s' completed.", self.__class__.__name__, str(self.output_dir))
            self.end(status="complete")
        elif issubclass(exc_type, KeyboardInterrupt):
            logger.warning("%s '%s' aborted.", self.__class__.__name__, str(self.output_dir))
            self.end(status="aborted")
            return False
        else:
            self.handle_exception(exc_type, exc_value, traceback)
            logger.warning("%s '%s' failed.", self.__class__.__name__, str(self.output_dir))
            self.end(status="failed")
            return False

    def synchronize(self):
        """Synchronize to existing output directory."""
        assert (
            self.metadata.output_dir is not None
        ), f"Cannot synchronize a {self.__class__.__name__} which has no `output_dir`."

        if self.metadata.output_dir.exists():
            metadata = self.load_metadata(self.metadata.output_dir)

            if not isinstance(self.metadata.configuration, dict):
                metadata.configuration = from_dict(type(self.metadata.configuration), metadata.configuration)

            self.metadata = metadata

            self.load_annotations()
        else:
            logger.warning("No metadata found. Skipping synchronization.")

        return self

    @classmethod
    def from_path(cls, path: PathLike, config_cls: Optional[Type] = None):
        metadata: Metadata = cls.load_metadata(path)

        experiment: Experiment
        if not metadata.is_series:
            if config_cls is not None:
                metadata.configuration = from_dict(config_cls, metadata.configuration)

            experiment = Trial(metadata)

        else:
            if config_cls is not None:
                metadata.configuration["base_config"] = from_dict(config_cls, metadata.configuration["base_config"])

            experiment = Series(metadata)

        experiment.load_annotations()

        return experiment

    @classmethod
    def all_from_path(cls, results_path: Union[str, PathLike], *, skip_hidden: bool = True) -> List["Experiment"]:
        """Load all experiments from the results_path."""
        results_path = Path(results_path)

        seen_dirs: Set[Path] = set()
        experiments = []

        for p in results_path.rglob("*/cordage.json"):
            if skip_hidden and any(part.startswith(".") for part in (p.relative_to(results_path)).parts):
                continue

            path = p.parent

            if path.parent in seen_dirs:
                # we already encountered a parent experiment (series)
                continue

            seen_dirs.add(path)

            try:
                experiments.append(cls.from_path(p.parent))
            except (JSONDecodeError, DaciteError) as exc:
                logger.warning("Couldn't load '%s': %s", str(path), str(exc))

        return sorted(experiments, key=lambda exp: exp.output_dir)

    def setup_log(self):
        logger = logging.getLogger()

        if not self.global_config.logging.use:
            return

        # in this case, a StreamHandler was set up by the series
        is_toplevel = self.metadata.parent_dir is None

        handler: logging.Handler

        if self.global_config.logging.to_stream and is_toplevel:
            # add colored stream handler
            format_str = "%(name)s:%(filename)s:%(lineno)d - %(message)s"

            if colorlog is not None:
                handler = colorlog.StreamHandler()
                handler.setFormatter(colorlog.ColoredFormatter(f"%(log_color)s%(levelname)-8s%(reset)s {format_str}"))
            else:
                handler = logging.StreamHandler()
                handler.setFormatter(logging.Formatter(f"%(levelname)-8s {format_str}"))

            logger.addHandler(handler)
            self.log_handlers.append(handler)

        if self.global_config.logging.to_file:
            # setup logging to local output_dir
            formatter = logging.Formatter("%(asctime)s %(levelname)-8s %(name)s:%(filename)s:%(lineno)d - %(message)s")
            handler = logging.FileHandler(self.log_path)
            handler.setFormatter(formatter)

            logger.addHandler(handler)
            self.log_handlers.append(handler)

    def teardown_log(self):
        logger = logging.getLogger()

        for handler in self.log_handlers:
            handler.close()
            logger.removeHandler(handler)


class Trial(Generic[T], Experiment):
    def __init__(
        self,
        metadata: Optional[Metadata] = None,
        /,
        config: Optional[T] = None,
        **kw,
    ):
        if metadata is not None:
            if len(kw) == 0 and config is None:
                super().__init__(metadata)
            else:
                msg = "If metadata are provided, config and additional keywords can not be set."
                raise TypeError(msg)
        else:
            super().__init__(configuration=config, **kw)

    @property
    def config(self):
        return self.metadata.configuration


class Series(Generic[T], Experiment):
    def __init__(
        self,
        metadata: Optional[Metadata] = None,
        /,
        base_config: Optional[T] = None,
        series_spec: Union[List[Dict], Dict[str, List], None] = None,
        series_skip: Optional[int] = None,
        **kw,
    ):
        if metadata is not None:
            assert len(kw) == 0 and base_config is None and series_spec is None
            super().__init__(metadata)
        else:
            if isinstance(series_spec, list):
                series_spec = [nest_items(flattened_items(trial_update, sep=".")) for trial_update in series_spec]

            super().__init__(
                configuration={"base_config": base_config, "series_spec": series_spec, "series_skip": series_skip}, **kw
            )

        self.validate_series_spec()

        self.trials: Optional[List[Trial[T]]] = None
        self.make_all_trials()

    def validate_series_spec(self):
        series_spec = self.series_spec

        if isinstance(series_spec, list):
            for config_update in series_spec:
                assert isinstance(config_update, dict)

        elif isinstance(series_spec, dict):

            def only_list_nodes(d):
                for v in d.values():
                    if isinstance(v, dict):
                        if not only_list_nodes(v):
                            return False
                    elif not isinstance(v, list):
                        return False
                    return True

            assert only_list_nodes(series_spec), f"Invalid series specification: {series_spec}"
        else:
            assert series_spec is None

    @property
    def base_config(self) -> T:
        return self.metadata.configuration["base_config"]

    @property
    def series_spec(self) -> Union[List[Dict], Dict[str, List], None]:
        return self.metadata.configuration["series_spec"]

    @property
    def series_skip(self) -> int:
        skip: Optional[int] = self.metadata.configuration.get("series_skip", None)

        if skip is None:
            return 0
        else:
            return skip

    @property
    def is_singular(self):
        return self.series_spec is None

    def __enter__(self):
        if not self.is_singular:
            super().__enter__()
        # else: do nothing

    def __exit__(self, *args):
        if not self.is_singular:
            super().__exit__(*args)
        # else: do nothing

    @overload
    def get_changing_fields(self, sep: Literal[None] = None) -> Set[Tuple[Any, ...]]:
        ...

    @overload
    def get_changing_fields(self, sep: str) -> Set[str]:
        ...

    def get_changing_fields(self, sep: Optional[str] = None) -> Union[Set[Tuple[Any, ...]], Set[str]]:
        keys: Set = set()

        if isinstance(self.series_spec, list):
            for trial_update in self.series_spec:
                for k, _ in flattened_items(trial_update, sep=sep):
                    keys.add(k)

        elif isinstance(self.series_spec, dict):
            for k, _ in flattened_items(self.series_spec, sep=sep):
                keys.add(k)

        return keys

    def get_trial_updates(self) -> Generator[Dict, None, None]:
        if isinstance(self.series_spec, list):
            yield from self.series_spec
        elif isinstance(self.series_spec, dict):
            keys, values = zip(*flattened_items(self.series_spec, sep="."))

            for update_values in product(*values):
                yield nest_items(zip(keys, update_values))
        else:
            yield {}

    def __len__(self):
        if isinstance(self.series_spec, list):
            assert self.trials is None or len(self.trials) == len(
                self.series_spec
            ), f"Number of existing ({len(self.trials)}) and expected trials ({len(self.series_spec)}) do not match."
            return len(self.series_spec)
        elif isinstance(self.series_spec, dict):
            num_trials = 1
            for _, values in flattened_items(self.series_spec):
                num_trials *= len(values)
            assert (
                self.trials is None or len(self.trials) == num_trials
            ), f"Number of existing ({len(self.trials)}) and expected trials ({num_trials}) do not match."
            return num_trials
        else:
            return 1

    def make_trial(self, **kw):
        additional_info = kw.pop("additional_info", None)

        fields_to_update: Dict[str, Any] = {
            "output_dir": None,
            "configuration": {},
            "additional_info": {},
            "status": "pending",
            "parent_dir": None,
            **kw,
        }

        trial_metadata = self.metadata.replace(**fields_to_update)

        if additional_info is not None:
            assert isinstance(additional_info, dict)
            trial_metadata.additional_info.update(additional_info)

        return Trial(trial_metadata)

    def make_all_trials(self):
        if self.series_spec is None:
            # single trial experiment
            logger.debug("Configuration yields a single experiment.")
            single_trial = self.make_trial(configuration=self.base_config)
            single_trial.annotations = self.annotations

            if self.metadata.output_dir:
                single_trial.metadata.output_dir = self.metadata.output_dir

            self.trials = [single_trial]

        else:
            logger.debug("The given configuration yields an experiment series with %d experiments.", len(self))
            self.trials = []

            for i, trial_update in enumerate(self.get_trial_updates()):
                trial_config_data: Dict[str, Any]

                if isinstance(self.base_config, dict):
                    trial_config_data = deepcopy(self.base_config)
                else:
                    trial_config_data = to_dict(self.base_config)

                logger.debug("Base configuration: %s", str(trial_config_data))
                logger.debug("Trial update: %s", str(trial_update))

                nested_update(trial_config_data, trial_update)

                trial_config: T
                if isinstance(self.base_config, dict):
                    trial_config = cast(T, trial_config_data)
                else:
                    trial_config = cast(T, from_dict(type(self.base_config), trial_config_data))

                if i < self.series_skip:
                    status = "skipped"
                else:
                    status = "pending"

                trial = self.make_trial(configuration=trial_config, additional_info={"trial_index": i}, status=status)
                self.trials.append(trial)

    def __iter__(self):
        return self.get_all_trials(include_skipped=False)

    def get_all_trials(self, *, include_skipped: bool = False) -> Generator[Trial[T], None, None]:
        assert self.trials is not None

        if not self.is_singular:
            skip = 0 if include_skipped else self.series_skip

            for i, trial in enumerate(self.trials[skip:], start=skip):
                trial_subdir = str(i).zfill(ceil(log10(len(self))))

                trial.metadata.output_dir = self.output_dir / trial_subdir

                yield trial
        else:
            assert len(self.trials) == 1
            yield self.trials[0]



import dataclasses
import logging
from datetime import datetime, timedelta
from os import PathLike
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

import dacite

logger = logging.getLogger("cordage")

T = TypeVar("T")


serialization_map: Dict[Type[Any], Callable[..., Any]] = {
    Path: str,
    datetime: datetime.isoformat,
    timedelta: lambda v: v.total_seconds(),
}

deserialization_map: Dict[Type[Any], Callable[..., Any]] = {
    datetime: lambda v: datetime.fromisoformat(v),
    timedelta: lambda v: timedelta(seconds=v),
}

types_to_cast: List[Type[Any]] = [Path, float, bool, int, str, tuple]


def get_loader(extension: str) -> Callable:
    """Load relevant module for reading a file with the given extension."""
    msg = f"Unrecognized file format: '.{extension}' (supported are .toml, .yaml, and .json)."
    if extension not in ("toml", "yaml", "yml", "yl", "json"):
        raise RuntimeError(msg)

    loader: Callable

    if extension == "toml":
        try:
            from toml import load as toml_loader

            loader = toml_loader
        except ModuleNotFoundError as exc:
            msg = f"Package toml is required to read .{extension} files."
            raise RuntimeError(msg) from exc

    elif extension in ("yaml", "yml", "yl"):
        try:
            from yaml import safe_load as yaml_loader

            loader = yaml_loader
        except ModuleNotFoundError as exc:
            msg = f"Package pyyaml is required to read .{extension} files."
            raise RuntimeError(msg) from exc
    else:
        try:
            from json import load as json_loader

            loader = json_loader
        except ModuleNotFoundError as exc:
            msg = f"Package json is required to read .{extension} files."
            raise RuntimeError(msg) from exc

    return loader


def read_dict_from_file(path: PathLike) -> Dict[str, Any]:
    """Read dictionary from toml, yaml, or json file.

    The file-type is inferred from the file extension.
    """
    extension = Path(path).suffix[1:]

    loader = get_loader(extension)

    with open(path, encoding="utf-8") as conf_file:
        return loader(conf_file)


def get_writer(extension: str) -> Callable:
    """Load relevant module for reading a file with the given extension."""
    if extension not in ("toml", "yaml", "yml", "yl", "json"):
        msg = f"Unrecognized file format: '.{extension}' (supported are .toml, .yaml, and .json)."
        raise RuntimeError(msg)

    writer: Callable

    if extension == "toml":
        try:
            from toml import dump as toml_dump

            writer = toml_dump
        except ModuleNotFoundError as exc:
            msg = f"Package toml is required to read .{extension} files."
            raise RuntimeError(msg) from exc

    elif extension in ("yaml", "yml", "yl"):
        try:
            from yaml import dump as yaml_dump

            writer = yaml_dump
        except ModuleNotFoundError as exc:
            msg = f"Package pyyaml is required to write .{extension} files."
            raise RuntimeError(msg) from exc
    else:
        try:
            from json import dump as json_dump

            writer = json_dump
        except ModuleNotFoundError as exc:
            msg = f"Package json is required to read .{extension} files."
            raise RuntimeError(msg) from exc

    return writer


def write_dict_to_file(path: PathLike, data: Mapping[str, Any]):
    """Write dictionary to toml, yaml, or json file.

    The file-type is inferred from the file extension.
    """
    extension = Path(path).suffix[1:]

    writer = get_writer(extension)

    with open(path, "w", encoding="utf-8") as conf_file:
        return writer(data, conf_file)


# no separator
@overload
def flattened_items(
    nested_dict: Dict[Any, Any], *, sep: Literal[None] = None, prefix: Tuple[Any, ...] = ()
) -> Generator[Tuple[Tuple[Any, ...], Any], None, None]:
    ...


# spearator given
@overload
def flattened_items(
    nested_dict: Dict[Any, Any], *, sep: str, prefix: Tuple[str, ...] = ()
) -> Generator[Tuple[str, Any], None, None]:
    ...


def flattened_items(
    nested_dict: Dict,
    *,
    sep: Optional[str] = None,
    prefix: Tuple[Any, ...] = (),
) -> Generator[Union[Tuple[str, Any], Tuple[Tuple[Any, ...], Any]], None, None]:
    """Iter over all items in a nested dictionary."""
    for k, v in nested_dict.items():
        flat_k: Tuple = (
            *prefix,
            k,
        )

        if isinstance(v, dict):
            for ik, iv in flattened_items(v, prefix=flat_k):
                if sep is None:
                    yield ik, iv
                else:
                    yield sep.join(ik), iv
        elif sep is None:
            yield flat_k, v
        else:
            yield sep.join(flat_k), v


def nested_update(target_dict: Dict, update_dict: Mapping):
    """Update a nested dictionary."""
    for k, v in update_dict.items():
        if isinstance(v, Mapping) and k in target_dict and isinstance(target_dict[k], dict):
            nested_update(target_dict[k], v)
        else:
            target_dict[k] = v

    return target_dict


def nest_items(flat_items: Iterable[Tuple[Union[str, Tuple[Any, ...]], Any]]) -> Dict[str, Any]:
    """Unflatten a dict.

    If any keys contain '.', sub-dicts will be created.
    """
    nested_dict: Dict[str, Any] = {}
    dicts_to_nest: List[str] = []

    for k, v in flat_items:
        k_tuple: Tuple[Any, ...]
        if isinstance(k, tuple):
            k_tuple = k
        else:
            # if key is of the form "a.b", split into tuple
            k_tuple = tuple(k.split("."))

        prefix = k_tuple[0]
        remainder = k_tuple[1:]

        if len(remainder) == 0:
            nested_dict[prefix] = v

        else:
            if prefix not in nested_dict:
                nested_dict[prefix] = {}
                dicts_to_nest.append(prefix)

            nested_dict[prefix][remainder] = v

    for k in dicts_to_nest:
        nested_dict[k] = nest_items(nested_dict[k].items())

    return nested_dict


def from_dict(data_class: Type[T], data: Mapping, *, strict: bool = True) -> T:
    config = dacite.Config(cast=types_to_cast, type_hooks=deserialization_map, strict=strict)
    return dacite.from_dict(data_class, data, config)


def from_file(config_cls: Type[T], path: PathLike, **kwargs) -> T:
    data: Mapping = read_dict_from_file(path)
    return from_dict(config_cls, data, **kwargs)


def apply_nested_type_mapping(data: Mapping, type_mapping: Mapping[Type, Callable]):
    result = {}

    for k, v in data.items():
        result[k] = v

        if isinstance(v, Mapping):
            result[k] = apply_nested_type_mapping(v, type_mapping)

        else:
            for t, func in type_mapping.items():
                if isinstance(v, t):
                    result[k] = func(v)
                    break

    return result


def get_nested_field(dataclass_instance, field_name: str) -> Any:
    assert dataclasses.is_dataclass(dataclass_instance)

    value = dataclass_instance

    for k in field_name.split("."):
        value = getattr(value, k)

    return value


def set_nested_field(dataclass_instance, field_name: str, value: Any):
    *first_keys, last_key = field_name.split(".")

    obj = dataclass_instance
    for k in first_keys:
        obj = getattr(obj, k)

    setattr(obj, last_key, value)


def to_dict(data: Any) -> dict:
    """Represent the fields and values of configuration as a (nested) dict."""
    mapping: Mapping

    if dataclasses.is_dataclass(data):
        mapping = dataclasses.asdict(data)
    else:
        mapping = cast(Mapping, data)

    return apply_nested_type_mapping(mapping, serialization_map)


def to_file(dataclass_instance, path: PathLike):
    """Write config to json, toml, or yaml file."""
    return write_dict_to_file(path, to_dict(dataclass_instance))



from typing import List

import pytest
from config_classes import NestedConfig as Config

import cordage
from cordage import Series


def test_trial_series_list(global_config, resources_path):
    trial_store: List[cordage.Trial] = []

    def func(config: Config, cordage_trial: cordage.Trial, trial_store=trial_store):  # noqa: ARG001
        trial_store.append(cordage_trial)

    config_file = resources_path / "series_list.yml"

    series = cordage.run(func, args=[str(config_file)], global_config=global_config)

    assert isinstance(series, Series)

    assert series.get_changing_fields() == {("alpha", "a"), ("alpha", "b"), ("beta", "a")}

    assert len(trial_store) == 3
    assert trial_store[0].config.alpha.a == 1
    assert trial_store[0].config.alpha.b == "b1"
    assert trial_store[0].config.beta.a == "c1"
    assert trial_store[0].metadata.additional_info["trial_index"] == 0
    assert trial_store[1].config.alpha.a == 2
    assert trial_store[1].config.alpha.b == "b2"
    assert trial_store[1].config.beta.a == "c2"
    assert trial_store[1].metadata.additional_info["trial_index"] == 1
    assert trial_store[2].config.alpha.a == 3
    assert trial_store[2].config.alpha.b == "b3"
    assert trial_store[2].config.beta.a == "c3"
    assert trial_store[2].metadata.additional_info["trial_index"] == 2

    for i, trial in enumerate(trial_store):
        assert trial.output_dir == global_config.base_output_dir / "experiment" / str(i)


@pytest.mark.parametrize("letter", "abc")
def test_more_trial_series(global_config, resources_path, letter):
    trial_store: List[cordage.Trial] = []

    def func(config: Config, cordage_trial: cordage.Trial, trial_store=trial_store):  # noqa: ARG001
        trial_store.append(cordage_trial)

    config_file = resources_path / f"series_{letter}.toml"

    cordage.run(func, args=[str(config_file), "--alpha.b", "b_incorrect"], global_config=global_config)

    assert len(trial_store) == trial_store[0].config.alphas * trial_store[0].config.betas

    for i, trial in enumerate(trial_store):
        assert trial.config.alpha.b == "b1"
        assert trial.config.beta.a == "c" + str(1 + (i // trial_store[0].config.alphas))
        assert trial.config.alpha.a == (1 + (i % trial_store[0].config.alphas))

        assert trial.metadata.parent_dir is not None
        assert trial.metadata.parent_dir.parts[-1] == "experiment"

        if len(trial_store) <= 10:
            assert trial.output_dir == global_config.base_output_dir / "experiment" / f"{i}"

        else:
            assert trial.output_dir == global_config.base_output_dir / "experiment" / f"{i:02}"


def test_invalid_trial_series(global_config, resources_path):
    def func(config: Config, cordage_trial: cordage.Trial):  # noqa: ARG001
        pass

    config_file = resources_path / "series_invalid.json"

    with pytest.raises(ValueError):
        cordage.run(func, args=[str(config_file)], global_config=global_config)

    assert (
        not global_config.base_output_dir.exists()
    ), "Since the configuration is invalid, the series should not be started and hence no output be created"


def test_trial_skipping(global_config, resources_path):
    trial_store: List[cordage.Trial] = []

    def func(config: Config, cordage_trial: cordage.Trial, trial_store=trial_store):  # noqa: ARG001
        trial_store.append(cordage_trial)

    config_file = resources_path / "series_list.yml"

    cordage.run(func, args=[str(config_file), "--series-skip", "1"], global_config=global_config)

    assert len(trial_store) == 2

    assert trial_store[0].metadata.additional_info["trial_index"] == 1
    assert trial_store[0].config.alpha.a == 2
    assert trial_store[0].config.alpha.b == "b2"
    assert trial_store[0].config.beta.a == "c2"

    assert trial_store[1].metadata.additional_info["trial_index"] == 2
    assert trial_store[1].config.alpha.a == 3
    assert trial_store[1].config.alpha.b == "b3"
    assert trial_store[1].config.beta.a == "c3"

    for i, trial in enumerate(trial_store, start=1):
        assert trial.output_dir == global_config.base_output_dir / "experiment" / str(i)



from dataclasses import dataclass
from pathlib import Path

import cordage


@dataclass
class Config:
    """config_description.

    :param a: a_help_str
    :param d: wrong_help_text
    """

    a: int
    b: str


def test_simple_func(global_config):
    def func(config: Config):
        assert isinstance(config, Config)
        assert config.a == 1
        assert config.b == "test"

    cordage.run(func, args=["--a", "1", "--b", "test"], global_config=global_config)


def test_func_with_output_dir_path(global_config):
    def func(config: Config, output_dir: Path):
        assert config.a == 1
        assert config.b == "test"
        assert isinstance(output_dir, Path)
        assert output_dir.exists()

    cordage.run(func, args=["--a", "1", "--b", "test"], global_config=global_config)


def test_func_with_output_dir_str(global_config):
    def func(config: Config, output_dir: str):
        assert config.a == 1
        assert config.b == "test"
        assert isinstance(output_dir, str)

    cordage.run(func, args=["--a", "1", "--b", "test"], global_config=global_config)


def test_func_with_trial(global_config):
    def func(config: Config, cordage_trial):
        assert config.a == 1
        assert config.b == "test"
        assert isinstance(cordage_trial, cordage.Trial)

        assert str(cordage_trial.output_dir) in repr(cordage_trial)

        assert cordage_trial.output_dir == cordage_trial.global_config.base_output_dir / "experiment"

    cordage.run(func, args=["--a", "1", "--b", "test"], global_config=global_config)


def test_explicit_config_class(global_config):
    def func(config):  # noqa: ARG001
        pass

    cordage.run(func, args=["--a", "1", "--b", "test"], global_config=global_config, config_cls=Config)



import logging
from typing import cast

import pytest
from config_classes import SimpleConfig as Config

import cordage
from cordage import Series


@pytest.mark.timeout(1)
def test_nested_trial_logging(global_config, capsys):
    def foo(config: Config):  # noqa: ARG001
        log = logging.getLogger("test-logger")

        log.warning("in_inner_trial")

    def bar(config: Config, cordage_trial):  # noqa: ARG001
        log = logging.getLogger("test-logger")

        log.warning("before_inner_trial")

        inner_trial = cordage.run(foo, args=[], global_config=global_config)

        log.warning("after_inner_trial")

        assert inner_trial.parent_dir == cordage_trial.output_dir

    cordage.run(bar, args=[], global_config=global_config)

    captured = capsys.readouterr()

    # 42 should appear exactly once in the resulting log output
    assert captured.err.count("in_inner_trial") == 1
    assert captured.err.find("before_inner_trial") < captured.err.find("in_inner_trial")
    assert captured.err.find("in_inner_trial") < captured.err.find("after_inner_trial")


@pytest.mark.timeout(1)
def test_nested_series_logging(global_config, capsys, resources_path):
    def foo(config: Config):
        log = logging.getLogger("test-logger")

        log.warning(f"in_inner_trial_{config.a}")

    def bar(config: Config, cordage_trial):  # noqa: ARG001
        log = logging.getLogger("test-logger")

        log.warning("before_inner_trial")

        inner_series = cordage.run(foo, args=[str(resources_path / "series_simple.yaml")], global_config=global_config)

        log.warning("after_inner_trial")

        assert inner_series.parent_dir == cordage_trial.output_dir
        for trial in cast(Series, inner_series):
            assert trial.parent_dir == inner_series.output_dir

    cordage.run(bar, args=[], global_config=global_config)

    captured = capsys.readouterr()

    # 42 should appear exactly once in the resulting log output
    assert captured.err.count("in_inner_trial") == 3
    assert captured.err.find("before_inner_trial") < captured.err.find("inner_trial_1")
    assert captured.err.find("in_inner_trial_1") < captured.err.find("in_inner_trial_2")
    assert captured.err.find("in_inner_trial_2") < captured.err.find("in_inner_trial_3")
    assert captured.err.find("in_inner_trial_3") < captured.err.find("after_inner_trial")



import argparse
import dataclasses
import inspect
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, ClassVar, Dict, List, Literal, Mapping, Optional, Type, Union, get_args, get_origin

from docstring_parser import parse as parse_docstring

from cordage.experiment import Experiment, Series, Trial
from cordage.global_config import GlobalConfig
from cordage.util import from_dict as config_from_dict
from cordage.util import logger, nest_items, nested_update, read_dict_from_file


class MissingType:
    def __repr__(self):
        return "<MISSING>"


MISSING = MissingType()


SUPPORTED_PRIMITIVES = (int, bool, str, float, Path)


class Singleton(type):
    _instances: ClassVar[Dict[Type, Any]] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class ExperimentStack(metaclass=Singleton):
    """Represents the stack of currently running experiments (necessary when running nested experiments).

    This class is used internally, to determine whether an experiment was started from within another.
    """

    def __init__(self):
        self.running: List[Experiment] = []

    def push(self, experiment: Experiment):
        """Push a new experiment on the stack."""
        self.running.append(experiment)

    def pop(self) -> Experiment:
        """Pop the currently running experiment from the stack."""
        return self.running.pop()

    def peek(self) -> Optional[Experiment]:
        """Get the currently active experiment from the stack (may be None)."""
        if len(self.running) > 0:
            return self.running[-1]
        else:
            return None

    def peek_dir(self) -> Optional[Path]:
        """Get the output_dir of the currently running experiment (may be None)."""
        if len(self.running) > 0:
            return self.running[-1].output_dir
        else:
            return None

    def __len__(self):
        return len(self.running)

    @contextmanager
    def with_experiment_on_stack(self, experiment: Experiment):
        """Put a new experiment on the stack and set its parent_dir to the experiment which was running so far."""
        if self.peek() == experiment:
            yield experiment

        else:
            experiment.metadata.parent_dir = self.peek_dir()
            self.push(experiment)
            try:
                with experiment:
                    yield experiment
            finally:
                self.pop()


experiment_stack: ExperimentStack = ExperimentStack()


class FunctionContext:
    """Wrapper for a function which accepts a dataclass configuration.

    This class can be used to:
    - parse argruments matching the config dataclass,
    - build a dictionary of the arguments expected by the function, and
    - call the function for each trial in a series.
    """

    global_config: GlobalConfig

    usage_str: str = "%(prog)s [-h] [config_file] <configuration options to overwrite>"

    def __init__(
        self,
        func: Callable,  # expects a dataclass
        global_config: GlobalConfig,
        description: Optional[str] = None,
        config_cls: Optional[Type] = None,
    ):
        self.global_config = global_config
        self.set_function(func)
        self.set_config_cls(config_cls)
        self.set_description(description)
        self.construct_argument_parser()

    def set_description(self, description: Optional[str] = None):
        if description is None:
            if self.func.__doc__ is not None:
                self.description = parse_docstring(self.func.__doc__).short_description
            else:
                self.description = self.func_name

        else:
            self.description = description

    def set_config_cls(self, config_cls: Optional[Type] = None):
        # derive configuration class
        if config_cls is None:
            self.main_config_cls = self.func_parameters[self.global_config.param_names.config].annotation

        else:
            self.main_config_cls = config_cls

        if not dataclasses.is_dataclass(self.main_config_cls):
            msg = (
                f"Configuration class could not be derived: Either pass a configuration dataclass via `config_cls` or"
                f"annotate the configuration parameter `{self.global_config.param_names.config}` with a dataclass."
            )
            raise TypeError(msg)

    @property
    def func(self) -> Callable:
        return self._func

    @property
    def func_parameters(self):
        return self._func_parameters

    @property
    def func_name(self):
        return self._func_name

    def set_function(self, func: Callable):
        self._func = func
        self._func_parameters = inspect.signature(func).parameters
        self._func_name = self.func.__name__

        if self.global_config.param_names.config not in self.func_parameters:
            msg = f"Callable must accept config argument (as `{self.global_config.param_names.config}`)."
            raise TypeError(msg)

    def construct_argument_parser(self):
        """Construct an argparser for a given config class."""
        # add parser arguments from dataclass

        self.argument_parser = argparse.ArgumentParser(description=self.description, usage=self.usage_str)

        self.add_arguments_to_parser(self.main_config_cls)

        self.argument_parser.add_argument(
            ".", metavar="config_file", nargs="?", help="Top-level config file to load.", type=Path, default=MISSING
        )

        self.argument_parser.add_argument(
            "--series-skip",
            type=int,
            metavar="n",
            help="Skip first n trials in the execution of a series.",
            default=MISSING,
            dest=self.global_config._series_skip_key,
        )

        if not self.global_config.config_only:
            self.argument_parser.add_argument(
                "--cordage-comment",
                type=str,
                help="Add a comment to the annotation of this series.",
                default=MISSING,
                dest=self.global_config._experiment_comment_key,
            )

            self.argument_parser.add_argument(
                "--output-dir",
                type=Path,
                help="Path to use as the output directory.",
                default=MISSING,
                dest=self.global_config._output_dir_key,
            )

    def _add_argument_to_parser(self, arg_name: str, arg_type: Any, help: str, **kw):  # noqa: A002
        # If the field is also a dataclass, recurse (nested config)
        if dataclasses.is_dataclass(arg_type):
            self.add_arguments_to_parser(arg_type, prefix=arg_name)

            self.argument_parser.add_argument(f"--{arg_name}", type=Path, default=MISSING, help=help, **kw)
        # Look the field annotation to determine which type of argument to add

        # Choice field
        elif get_origin(arg_type) is Literal:
            # Value must be from this set
            choices = get_args(arg_type)

            literal_arg_type = type(choices[0])

            if any(not isinstance(c, literal_arg_type) for c in choices):
                msg = f"If Literal is used, all values must be of the same type ({arg_name})."
                raise TypeError(msg)

            self.argument_parser.add_argument(
                f"--{arg_name}", type=literal_arg_type, choices=choices, default=MISSING, help=help, **kw
            )

        elif get_origin(arg_type) is Union:
            args = [arg for arg in get_args(arg_type) if arg != type(None)]

            if len(args) == 1:
                # optional
                self._add_argument_to_parser(arg_name, args[0], help=help, **kw)

            else:
                msg = "Config parser does not support Union annotations with more than one type other than None."
                raise TypeError(msg)

        # Boolean field
        elif arg_type == bool:
            # Create a true and a false flag -> the destination is identical
            self.argument_parser.add_argument(
                f"--{arg_name}", action="store_true", default=MISSING, help=help + " (set the value to True)", **kw
            )

            self.argument_parser.add_argument(
                f"--not-{arg_name}",
                dest=arg_name,
                action="store_false",
                default=MISSING,
                help=help + " (set the value to False)",
                **kw,
            )

        elif arg_type in SUPPORTED_PRIMITIVES:
            self.argument_parser.add_argument(f"--{arg_name}", type=arg_type, default=MISSING, help=help, **kw)

    def add_arguments_to_parser(self, config_cls: Type, prefix: Optional[str] = None):
        """Recursively (if nested) iterate over all fields in the dataclass and add arguments to parser."""

        # read documentation of config dataclass. If no help metadata is given, this will be used a the help text.
        param_doc = {}
        if config_cls.__doc__ is not None:
            # parse doc text, to generate help text for fields
            for param in parse_docstring(config_cls.__doc__).params:
                param_doc[param.arg_name] = param.description

        # Iterate over all fields in the dataclass to add arguments to the parser
        for field in dataclasses.fields(config_cls):
            # Set prefixed argument name
            if prefix is not None:
                arg_name = f"{prefix}.{field.name}"
            else:
                arg_name = field.name

            # Retrieve help text
            help_text = field.metadata.get("help", param_doc.get(field.name, ""))

            self._add_argument_to_parser(arg_name, field.type, help=help_text)

    def remove_missing_values(self, data: Mapping) -> Dict[str, Any]:
        return {k: v for k, v in data.items() if v is not MISSING}

    def construct_func_kwargs(self, trial: Trial):
        # construct arguments for the passed callable
        func_kw: Dict[str, Any] = {}

        # check if any other parameters are expected which can be resolved
        for name, param in self.func_parameters.items():
            assert param.kind != param.POSITIONAL_ONLY, "Cordage currently does not support positional only parameters."

            if name == self.global_config.param_names.config:
                # pass the configuration
                func_kw[name] = trial.config

            elif not self.global_config.config_only:
                if name == self.global_config.param_names.output_dir:
                    # pass path to output directory
                    if issubclass(param.annotation, str):
                        func_kw[name] = str(trial.output_dir)
                    else:
                        func_kw[name] = trial.output_dir

                elif name == self.global_config.param_names.trial_object:
                    # pass trial object
                    func_kw[name] = trial

        return func_kw

    def parse_args(self, args: Optional[List[str]] = None) -> Experiment:
        """Parse the command line arguments.

        :param args: A list of arguments (usually from the CLI). If none, sys.argv[1:] is used.

        :return: the resulting experiment (the resulting configuration of the experiment) is the result of the default
            values, a potentially loaded config file, and parameters passed via the CLI args.
        """

        if args is None:
            # args default to the system args
            args = sys.argv[1:]
        else:
            args = list(args)

        # construct parser
        argument_data: dict = vars(self.argument_parser.parse_args(args))
        argument_data = self.remove_missing_values(argument_data)

        conf_file_comment: Optional[str] = None
        cli_series_comment: Optional[str] = None

        series_kw = {
            "function": self.func_name,
            "global_config": self.global_config,
            "additional_info": {"description": self.description, "parsed_arguments": args},
        }

        if not self.global_config.config_only:
            cli_series_comment = argument_data.pop(self.global_config._experiment_comment_key, None)
            series_kw["output_dir"] = argument_data.pop(self.global_config._output_dir_key, None)

        config_path = argument_data.pop(".", None)

        argument_data = nest_items(argument_data.items())

        if config_path is not None:
            new_conf_data = read_dict_from_file(config_path)

            new_conf_data = nest_items(new_conf_data.items())

            series_kw["series_spec"] = new_conf_data.pop(self.global_config._series_spec_key, None)

            nested_update(new_conf_data, argument_data)

            argument_data = new_conf_data

            if not self.global_config.config_only:
                # another series comment might be given via the config file
                # in this case, the comments are added to another
                conf_file_comment = argument_data.pop(self.global_config._experiment_comment_key, None)
        else:
            series_kw["series_spec"] = None

        # series skip might be given via the command line ("--series-skip <n>") or a config file "__series-skip__"
        series_kw["series_skip"] = argument_data.pop(self.global_config._series_skip_key, None)

        series_kw["base_config"] = config_from_dict(
            self.main_config_cls, argument_data, strict=self.global_config.strict_mode
        )

        series: Series = Series(**series_kw)

        if not self.global_config.config_only:
            if cli_series_comment is not None and conf_file_comment is not None:
                series.comment = conf_file_comment + "\n\n" + cli_series_comment
            else:
                series.comment = conf_file_comment or cli_series_comment or ""

        logger.debug("%d experiments found in configuration", len(series))

        if series.is_singular:
            return next(iter(series))
        else:
            return series

    def from_configuration(
        self,
        config=None,
        base_config=None,
        series_spec=None,
        series_skip: Optional[int] = None,
        comment: Optional[str] = None,
    ) -> Experiment:
        _usage = "Either pass `config` or `base_config` and `series_spec`"

        if config is not None:
            assert base_config is None and series_spec is None and series_skip is None, _usage

            trial: Trial = Trial(
                function=self.func_name,
                config=config,
                global_config=self.global_config,
                additional_info={"description": self.description},
            )

            trial.comment = comment

            return trial

        else:
            assert base_config is not None and series_spec is not None

            series: Series = Series(
                function=self.func_name,
                base_config=base_config,
                global_config=self.global_config,
                series_spec=series_spec,
                series_skip=series_skip,
                additional_info={"description": self.description},
            )
            series.comment = comment
            return series

    def execute(self, experiment: Experiment):
        """Execute a given experiment (with the function of this `FunctionContext`)."""
        if isinstance(experiment, Trial):
            if self.global_config.config_only:
                # execute function with the constructed keyword arguments
                func_kw = self.construct_func_kwargs(experiment)
                experiment.metadata.result = self.func(**func_kw)

            else:
                # only use stack if full feature-set is used
                with experiment_stack.with_experiment_on_stack(experiment):
                    func_kw = self.construct_func_kwargs(experiment)
                    experiment.metadata.result = self.func(**func_kw)

        elif isinstance(experiment, Series):
            if self.global_config.config_only:
                for trial in experiment:
                    self.execute(trial)
            else:
                with experiment_stack.with_experiment_on_stack(experiment):
                    for trial in experiment:
                        self.execute(trial)

        else:
            msg = "Passed object must be Trial or Series"
            raise TypeError(msg)



from dataclasses import dataclass, field

import pytest

import cordage


@dataclass
class DataConfig:
    """Description of the data config.

    :param name: Name of the dataset
    """

    name: str
    version: int


@dataclass
class HyperParameterConfig:
    learning_rate: float = 5e-5
    weight_decay: float = 0.0


@dataclass
class NestedConfig:
    data: DataConfig
    hyper_params: HyperParameterConfig = field(default_factory=HyperParameterConfig)


def test_nested_config(global_config):
    def func(config: NestedConfig):
        assert config == NestedConfig(
            data=DataConfig(name="mnist", version=1),
            hyper_params=HyperParameterConfig(learning_rate=2.0, weight_decay=0.0),
        )

    cordage.run(
        func,
        args=["--data.name", "mnist", "--data.version", "1", "--hyper_params.learning_rate", "2"],
        global_config=global_config,
    )


def test_nested_config_overwrite(global_config, resources_path):
    def func(config: NestedConfig):
        assert config == NestedConfig(
            data=DataConfig(name="mnist", version=1),
            hyper_params=HyperParameterConfig(learning_rate=2.0, weight_decay=0.0),
        )

    config_file = resources_path / "nested_c.json"

    cordage.run(
        func,
        args=[str(config_file), "--data.name", "mnist", "--data.version", "1", "--hyper_params.learning_rate", "2"],
        global_config=global_config,
    )


def test_nested_loading(global_config, resources_path):
    def func(config: NestedConfig):
        assert config.data.name == "mnist"
        assert config.data.version == 1
        assert isinstance(config.hyper_params.learning_rate, float)
        assert config.hyper_params.learning_rate == 2.0
        assert config.hyper_params.weight_decay == 0.0

    config_file = resources_path / "nested_a.json"

    cordage.run(
        func,
        args=[str(config_file)],
        global_config=global_config,
    )


def test_mixed_nested_loading(global_config, resources_path):
    def func(config: NestedConfig):
        assert config.data.name == "mnist"
        assert config.data.version == 2
        assert isinstance(config.hyper_params.learning_rate, float)
        assert config.hyper_params.learning_rate == 2.0
        assert config.hyper_params.weight_decay == 0.0

    config_file = resources_path / "nested_b.yaml"

    cordage.run(
        func,
        args=[str(config_file), "--data.version", "2"],
        global_config=global_config,
    )


def test_additional_keys_exception(global_config, resources_path):
    def func(config: NestedConfig):  # noqa: ARG001
        pass

    with pytest.raises(cordage.UnexpectedDataError):
        cordage.run(func, args=[str(resources_path / "nested_add_key.json")], global_config=global_config)

    # with strict mode off, this should not raise an exception
    global_config.strict_mode = False
    cordage.run(func, args=[str(resources_path / "nested_add_key.json")], global_config=global_config)



from dataclasses import dataclass
from pathlib import Path

import cordage


@dataclass
class Config:
    lr: float = 5e-5
    name: str = "MNIST"


def train(config: Config, output_dir: Path):
    """Help text which will be shown."""

    print(output_dir)  # noqa: T201
    print(config)  # noqa: T201


if __name__ == "__main__":
    cordage.run(train)



from time import sleep
from typing import List

import pytest
from config_classes import SimpleConfig

import cordage
from cordage import Experiment, FunctionContext, Series, Trial


@pytest.mark.timeout(2)
def test_timing(global_config):
    def func(config: SimpleConfig, cordage_trial: cordage.Trial):  # noqa: ARG001
        sleep(1)

    trial = cordage.run(func, args=[], global_config=global_config)

    assert trial.metadata.duration.total_seconds() < 1.1
    assert trial.metadata.duration.total_seconds() > 0.9
    assert trial.metadata.status == "complete"

    metadata_path = trial.output_dir / "cordage.json"

    assert trial.metadata_path == metadata_path
    assert metadata_path.exists()

    experiment = Experiment.from_path(metadata_path)
    metadata = experiment.metadata

    assert metadata.duration.total_seconds() < 1.1
    assert metadata.duration.total_seconds() > 0.9
    assert metadata.status == "complete"


@pytest.mark.timeout(60)
def test_trial_id_collision(global_config):
    trial_store: List[cordage.Trial] = []

    def func(config: SimpleConfig, cordage_trial: cordage.Trial, trial_store=trial_store):  # noqa: ARG001
        trial_store.append(cordage_trial)

    for _ in range(1010):
        cordage.run(func, args=[], global_config=global_config)

    assert str(trial_store[7].output_dir).endswith("_08")
    assert str(trial_store[42].output_dir).endswith("_43")
    assert str(trial_store[123].output_dir).endswith("__0124")
    assert str(trial_store[1009].output_dir).endswith("__1010")


@pytest.mark.timeout(1)
def test_function_context_from_configuration(global_config):
    def func(config: SimpleConfig):  # noqa: ARG001
        pass

    context = FunctionContext(func, global_config=global_config)

    trial = context.from_configuration(config=SimpleConfig())

    assert isinstance(trial, Trial)

    series = context.from_configuration(base_config=SimpleConfig(), series_spec={"a": [1, 2, 3]})

    assert isinstance(series, Series)

    assert len(series) == 3


@pytest.mark.timeout(1)
def test_output_dir_path_correction(monkeypatch, tmp_path):
    def func(config: SimpleConfig):  # noqa: ARG001
        pass

    run_dir = tmp_path / "run"
    run_dir.mkdir()
    monkeypatch.chdir(run_dir)

    exp = cordage.run(func, args=[])

    output_dir = exp.output_dir.resolve()

    test_dir = tmp_path / "test"
    test_dir.mkdir()
    monkeypatch.chdir(test_dir)

    all_exp = Experiment.all_from_path("../run/results")

    assert len(all_exp) == 1
    assert str(all_exp[0].output_dir).startswith("..")
    assert all_exp[0].output_dir.resolve() == output_dir
    assert all_exp[0].result is None


def test_config_only(global_config):
    def func(config: SimpleConfig):  # noqa: ARG001
        pass

    cordage.run(func, args=[], global_config=global_config, config_only=True)

    # since no outputdir is assigned, the base dir should not be created
    assert not global_config.base_output_dir.exists()


def test_config_only_func_params(global_config):
    def func(config: SimpleConfig, output_dir):  # noqa: ARG001
        pass

    with pytest.raises(TypeError):
        # function should not need an output_dir
        cordage.run(func, args=[], global_config=global_config, config_only=True)
